//
// stund/stund.cxx
// 
// Copyright (c) 2002 Alan Hawrylyshen
// 
/// @author Alan Hawrylyshen
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the
// "Software"), to deal in the Software without restriction, including
// without limitation the rights to use, copy, modify, merge, publish,
// distribute, sublicense, and/or sell copies of the Software, and to
// permit persons to whom the Software is furnished to do so, subject to
// the following conditions:
// 
// The above copyright notice and this permission notice shall be
// included in all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
// TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
// SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//


#include <cassert>
#include <iostream>
#include <stdlib.h>
#include <unistd.h>

#include "util.h"
#include "udp.h"
#include "stun.h"

static struct 
{
      StunAddress myAddr;
      StunAddress altAddr;
      bool verbose;
      bool background;
} options;


/// usage() - Print usage message
void
usage()
{
   cerr << "usage: stund [-v] -a alternateServer -o otherport -h hostInterface -p port [-b]" << endl;
}


/**
  stund - STUN server.
  @param -v Verbose.
  @param -b run in background
  @param -a Address of the alternate STUN server.
  @param -p The port number to use as the base port.
  @param -h The address of this STUN server.
  @param -o The port number to use as an alternate port.
   Defaults to base-port + 1.
 
  OR

   @param --verbose Verbose
   @param --alternate-server Address of the alternate STUN server.
   @param --port The base port number.
   @param --server Address of this server.
   @param --alternate-port The alternate port to use (defaults to base + 1 ).
 */

int
main(int argc, char* argv[])
{
    assert( sizeof(UInt8 ) == 1 );
    assert( sizeof(UInt16) == 2 );
    assert( sizeof(UInt32) == 4 );

    while (true)
    {

       static struct option long_options[7] = 
       {
          {"other-port", 1, 0, 'o'},
          {"alternate-server",1, 0, 'a'},
          {"port", 1, 0, 'p'},
          {"server", 1, 0, 'h'},
          {"verbose", 0, 0, 'v'},
          {"background", 0, 0, 'b'},
          {0,0,0,0}
       };

       int c;
       int option_index = 0;
       unsigned short junk;
       
       c = local_getopt_long(argc,argv,"a:h:o:p:vb",long_options,&option_index);

       if (c == -1) break;

       switch (c)
       {
          case 'h':
             parseHostName(optarg,
                           &options.myAddr.addr.v4addr,
                           &junk);
             break;

          case 'p':
             options.myAddr.addrHdr.port = strtol(optarg, NULL, 0);
             break;

          case 'o':
             options.altAddr.addrHdr.port = strtol(optarg,NULL,0);
             break;

          case 'v':
             options.verbose = true;
             break;

          case 'b':
             options.background = true;
             break;

          case 'a':
             parseHostName(optarg,
                           &options.altAddr.addr.v4addr,
                           &junk);
             break;

          default:
             usage();
             exit(1);
       }

    }

    if ( options.verbose )
    {
       clog << "Running with on interface "
            << options.myAddr << " with alternate " 
            << options.altAddr << endl;
    }
    
    options.myAddr.addrHdr.family  = IPv4Family;
    options.altAddr.addrHdr.family = IPv4Family;
    
    if (( options.myAddr.addrHdr.port  == 0) ||
        ( options.altAddr.addrHdr.port == 0) ||
        ( options.myAddr.addr.v4addr  == 0 ) ||
        ( options.altAddr.addr.v4addr == 0 ) )
    {
       usage();
       exit(1);
    }
    
    initNetwork();

#if defined(WIN32)

    if (  options.background )
    {
        cerr << "The -b background option does not work in windows" << endl;
    }

    stunServer( options.myAddr, options.altAddr, options.verbose);
#else
    if ( options.background )
    {
       pid_t pid = fork();

       if (pid < 0)
       {
          cerr << "fork: unable to fork" << endl;
          return -1;
       }

       if (pid == 0) //child
       {
          // Should dissociate from terminal etc here....
          stunServer( options.myAddr, options.altAddr, options.verbose);
          // NOTREACHED
       }
    }
    else
    {
       stunServer( options.myAddr, options.altAddr, options.verbose);
    }
#endif

    return 0;
}


// Local Variables:
// mode:c++
// c-file-style:"ellemtel"
// c-file-offsets:((case-label . +))
// indent-tabs-mode:nil
// End:
